#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Extract Tokens using a Sliding Window Algorithm """


from pprint import pformat

from baseblock import Stopwatch
from baseblock import BaseObject


class SlidingWindowExtract(BaseObject):
    """ Extract Tokens using a Sliding Window Algorithm """

    def __init__(self,
                 tokens: list,
                 gram_size: int):
        """
        Created:
            24-Aug-2022
            craigtrim@gmail.com
        """
        BaseObject.__init__(self, __name__)
        self._tokens = tokens
        self._gram_size = gram_size

    def _process(self) -> list:
        results = []
        tokens = self._tokens

        if self._gram_size == len(tokens):
            return [tokens]

        if self._gram_size == 1:
            return [[x] for x in tokens]

        x = 0
        y = x + self._gram_size

        while y <= len(tokens):
            results.append(tokens[x: y])

            x += 1
            y = x + self._gram_size

        results = [' '.join(x) for x in results]

        return results

    def process(self) -> list:
        sw = Stopwatch()

        results = self._process()

        if self.isEnabledForDebug:

            self.logger.debug('\n'.join([
                "Sliding Window Extract Completed",
                f"\tGram Size: {self._gram_size}",
                f"\tTotal Tokens: {len(results)}",
                f"\tTotal Time: {str(sw)}"]))

            if self._tokens != results:
                self.logger.debug('\n'.join([
                    "Sliding Window Extract Results",
                    f"\tTokens: {pformat(results, indent=4)}"]))

        return results



from .bp import *
from .svc import *
from .dmo import *
from .dto import *


from .bp.tagger import Tagger

tagapi = Tagger()


def dbpedia_tag(input_text) -> dict:
    return tagapi.process(input_text)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Perform Sliding Window Extraction for Candidate Synonym Swapping """


from typing import Callable

from baseblock import Enforcer
from baseblock import Stopwatch
from baseblock import BaseObject

from dbpedia_tag.dmo import SlidingWindowExtract
from dbpedia_tag.dmo import SlidingWindowBlacklist
from dbpedia_tag.dmo import SlidingWindowValidate


class ExactMatchFinder(BaseObject):
    """ Perform Sliding Window Extraction for Candidate Synonym Swapping """

    def __init__(self,
                 gram_size: int,
                 entity_exists: Callable):
        """
        Created:
            24-Aug-2022
            craigtrim@gmail.com
        """
        BaseObject.__init__(self, __name__)
        self._gram_size = gram_size
        self._entity_exists = entity_exists

    def _process(self,
                 tokens: list) -> dict:

        candidates_1 = SlidingWindowExtract(
            tokens=tokens,
            gram_size=self._gram_size).process()

        if not candidates_1 or not len(candidates_1):
            return None

        candidates_2 = SlidingWindowBlacklist().process(candidates_1)

        candidates_3 = SlidingWindowValidate(
            candidates=candidates_2,
            entity_exists=self._entity_exists).process()

        unmatched = [x for x in candidates_1 if x not in candidates_3]
        if not unmatched or not len(unmatched):
            return None

        return {
            'input': tokens,
            'extracted': candidates_1,
            'validated': candidates_3,
            'unmatched': unmatched
        }

    def process(self,
                tokens: list) -> list:
        sw = Stopwatch()

        d_results = self._process(tokens)

        if self.isEnabledForInfo:

            def total_results() -> int:
                if d_results:
                    return len(d_results)
                return 0

            if self.isEnabledForInfo:
                self.logger.info('\n'.join([
                    "Sliding Window Completed",
                    f"\tGram Size: {self._gram_size}",
                    f"\tTotal Results: {total_results()}",
                    f"\tTotal Time: {str(sw)}"]))

        if self.isEnabledForDebug and d_results:
            Enforcer.keys(d_results, 'input', 'extracted',
                          'validated', 'unmatched')

        return d_results



from .sliding_window_extract import SlidingWindowExtract
from .sliding_window_validate import SlidingWindowValidate
from .sliding_window_blacklist import SlidingWindowBlacklist
from .candidate_fuzzy_match import CandidateFuzzyMatch



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Perform Token Candidate Fuzzy Matching """


from enum import Enum
from typing import Callable
from collections import defaultdict

from baseblock import Stopwatch
from baseblock import TextUtils
from baseblock import BaseObject


class MatchType(Enum):  # debug codes
    TYPE_00 = 0
    TYPE_10 = 10
    TYPE_11 = 11
    TYPE_12 = 12
    TYPE_20 = 20
    TYPE_21 = 21
    TYPE_30 = 30
    TYPE_31 = 31
    TYPE_40 = 40
    TYPE_41 = 41
    TYPE_50 = 50
    TYPE_51 = 51


class CandidateFuzzyMatch(BaseObject):
    """ Perform Token Candidate Fuzzy Matching """

    def __init__(self,
                 entity_finder: Callable):
        """
        Created:
            24-Aug-2022
            craigtrim@gmail.com
        """
        BaseObject.__init__(self, __name__)
        self._entity_finder = entity_finder

    @staticmethod
    def _trie_exclusion(token: str,
                        candidates: list) -> list:
        """ Progressively narrow down tokens

        Args:
            token (str): the token to match
            candidates (list): the incoming candidates

        Returns:
            list: the remaining candidates
        """

        token = token.replace(' ', '_').lower()

        for i in range(4, len(token) - 1):

            if len(token) < i:
                return candidates
            if len(candidates) <= 3:
                return candidates

            candidates = [candidate for candidate in candidates
                          if i < len(candidate) and
                          candidate[i-1] == token[i-1]]

            if len(candidates) == 1:
                score = TextUtils.jaccard_similarity(token, candidates[0])
                if score >= 0.90:
                    return candidates
                return []

        return candidates

    @staticmethod
    def _length_filter(token: str,
                       candidates: list) -> list:
        """ Exclude Candidates to Short or too Long to be useful matches

        Heuristic:
            Narrow candidates down to +/- 2
            this covers suffixes such as the gerund form (e.g., "<word>ing" => "<word>")

        Args:
            token (str): the token to match
            candidates (list): the incoming candidates

        Returns:
            list: the remaining candidates
        """
        x = len(token) - 3
        y = len(token) + 3
        return [candidate for candidate in candidates
                if len(candidate) >= x and len(candidate) <= y]

    @staticmethod
    def _suffix_match(token: str,
                      candidates: list) -> str or None:
        """ Search Candidate for Common Suffixes

        e.g.,   if the token is 'Restaurants' and there exists a candidate 'Restaurant'
                then use this

        Args:
            token (str): the token to match
            candidates (list): the incoming candidates

        Returns:
            str: the closest candidate (if any)
        """

        suffixes = ['s', 'ies', 'ing', 'ed', 'ly']
        for suffix in suffixes:
            if token.endswith(suffix):
                candidates = [candidate for candidate in candidates
                              if token[:-len(suffix)] == candidate]
                if len(candidates) == 1:
                    return candidates[0]

    def _similarity_match(self,
                          token: str,
                          candidates: list) -> str or None:
        """ Use Jaccard Similarity to find the closest match

        Args:
            token (str): the token to match
            candidates (list): the incoming candidates

        Returns:
            str: the closest candidate (if any)
        """

        d_score = defaultdict(list)
        for candidate in candidates:
            token = token.replace(' ', '_').lower()
            score = TextUtils.jaccard_similarity(token, candidate)

            if self.isEnabledForDebug:
                self.logger.debug('\n'.join([
                    "Similarity Score Completed",
                    f"\tToken: {token}",
                    f"\tCandidate: {candidate}",
                    f"\tScore: {score}"]))

            if score >= 0.90:
                d_score[score].append(candidate)

        if not len(d_score):
            return None

        # find the values with the highest score
        # there may be more than 1
        values = d_score[max(d_score)]

        # find the values with the minimum length
        # it seems unlikely there would be more than 1, since this would imply duplicate values
        d_values = {len(value): value for value in values}

        return d_values[min(d_values)]

    def _process(self,
                 token: str) -> tuple or None:

        if len(token) < 3:
            return None, MatchType.TYPE_00.name

        candidates = self._entity_finder(token)

        if not candidates or not len(candidates):
            return None, MatchType.TYPE_10.name
        if token in candidates:
            return token, MatchType.TYPE_11.name
        if len(candidates) == 1:
            return candidates[0], MatchType.TYPE_12.name

        candidates = self._length_filter(token, candidates)
        if not len(candidates):
            return None, MatchType.TYPE_20.name
        if len(candidates) == 1:
            return candidates[0], MatchType.TYPE_21.name

        candidates = self._trie_exclusion(token, candidates)
        if not len(candidates):
            return None, MatchType.TYPE_30.name
        if len(candidates) == 1:
            return candidates[0], MatchType.TYPE_31.name

        candidate = self._suffix_match(token, candidates)
        if candidate:
            return candidate, MatchType.TYPE_40.name

        candidate = self._similarity_match(token, candidates)
        if candidate:
            return candidate, MatchType.TYPE_50.name

        return None, MatchType.TYPE_51.name

    def process(self,
                token: str) -> str or None:
        sw = Stopwatch()

        result, reason = self._process(token)

        if self.isEnabledForDebug:
            self.logger.debug('\n'.join([
                "Fuzzy Match Completed",
                f"\tUnknown Token: {token}",
                f"\tFuzzy Token: {result}",
                f"\tMatch Type: {reason}",
                f"\tTotal Time: {str(sw)}"]))

        return result
